#include <stdio.h>
#include <string.h>
#include <stdlib.h>

FILE *openSkipHeader(const char *path, FILE *fpw, char verbose)
{
	FILE *fpr = fopen(path, "rb");
	int c;
	if (fpr == NULL) {
		puts("fopen error");
		exit(1);
	}
	if (verbose != 0)
		printf("format: ");
	do {
		c = fgetc(fpr);
		if (c == EOF) break;
		if (fpw != NULL)
			fputc(c, fpw);
		if (c != '\0' && verbose != 0)
			putchar(c);
	} while (c != '\0');
	if (verbose != 0)
		printf("\nnote: ");
	do {
		c = fgetc(fpr);
		if (c == EOF) break;
		if (fpw != NULL)
			fputc(c, fpw);
		if (c != '\0' && verbose != 0)
			putchar(c);
	} while (c != '\0');
	if (verbose != 0)
		putchar('\n');
	return fpr;
}

void getSizFlgId(FILE *fp, int *siz, int *flg, char *id)
{
	int i = 0, c;
	*siz = 0;
	*flg = 0;
	*id = '\0';
	fread(siz, 1, 4, fp);
	if (*siz == 0) goto fin;
	fread(flg, 1, 4, fp);
	do {
		c = fgetc(fp);
		if (c == EOF) {
			puts("format error");
			exit(1);
		}
		if (i >= 255 && c != '\0') {
			printf("too long id: %.64s\n", id);
			exit(1);
		}
		id[i++] = c;
	} while (c != '\0');
fin:
	return;
}

void putSizFlgId(FILE *fp, int siz, int flg, const char *id)
{
	fwrite(&siz,  1, 4, fp);
	fwrite(&flg, 1, 4, fp);
	fwrite(id, 1, strlen(id) + 1, fp);
	return;
}

void copyBytes(FILE *fpr, FILE *fpw, int siz)
{
	int i, c;
	for (i = 0; i < siz; i++) {
		c = fgetc(fpr);
		if (c == EOF) break;
		if (fpw != NULL)
			fputc(c, fpw);
	}
	return;
}

void putSiz0(FILE *fp)
{
	int i = 0;
	fwrite(&i, 1, 4, fp);
	fclose(fp);
	return;
}

void searchId(FILE *fp, const char *id, int *siz, int *flg)
{
	int i, c;
	char tid[256];
	for (;;) {
		getSizFlgId(fp, siz, flg, tid);
		if (*siz == 0) break;
		if (strcmp(tid, id) == 0) break;
		copyBytes(fp, NULL, *siz);
	}
	return;
}

int deleteId(const char *path, FILE *fpw, const char *id)
{
	FILE *fpr = openSkipHeader(path, fpw, 0);
	char tid[256];
	int i, siz, flg, fnd = 0, c;
	for (;;) {
		getSizFlgId(fpr, &siz, &flg, tid);
		if (siz == 0) break;
		if (fnd == 0 && strcmp(tid, id) == 0) {
			fnd++;
			copyBytes(fpr, NULL, siz);
		} else {
			putSizFlgId(fpw, siz, flg, tid);
			copyBytes(fpr, fpw, siz);
		}
	}
	fclose(fpr);
	/* Ōsiz=0͂ďo͂ĂȂ! */
	return fnd;
}

void renameTmp(FILE *fp, const char *s, const char *d)
{
	if (fp != NULL)
		putSiz0(fp);
	if (rename(s, d) != 0) {
		remove(d);
		rename(s, d);
	}
	return;
}

int main(int argc, const char **argv)
{
	int c, siz, flg, i, cc, fnd = 0;
	char id[256];

	if (argc <= 2) {
		printf("prompt>pdbtol list   (pdb-file)\n");
		printf("prompt>pdbtol dump   (pdb-file)\n");
		printf("prompt>pdbtol string (pdb-file) (var-id)\n");
		printf("prompt>pdbtol getint (pdb-file) (var-id)\n");
		printf("prompt>pdbtol setint (pdb-file) (var-id) (integer)\n");
		printf("prompt>pdbtol new    (pdb-file) (var-id) (size)\n");
		printf("prompt>pdbtol resize (pdb-file) (var-id) (new-size)\n");
		printf("prompt>pdbtol delete (pdb-file) (var-id)\n");
		printf("prompt>pdbtol rename (pdb-file) (old-id) (new-id)\n");
		exit(0);
	}

	if (strcmp(argv[1], "list") == 0) {
		FILE *fp = openSkipHeader(argv[2], NULL, 1);
		putchar('\n');
		for (;;) {
			getSizFlgId(fp, &siz, &flg, id);
			if (siz == 0) break;
			printf("%s (%dbytes)\n", id, siz);
			for (i = 0; i < siz; i++) {
				c = fgetc(fp);
				if (c == EOF) break;
			}
		}
		fclose(fp);
	}

	if (strcmp(argv[1], "dump") == 0) {
		FILE *fp = openSkipHeader(argv[2], NULL, 1);
		for (;;) {
			putchar('\n');
			getSizFlgId(fp, &siz, &flg, id);
			if (siz == 0) break;
			printf("%s (%dbytes) ", id, siz);
			if (flg != 0)
				printf("(flags=0x%08x) ", flg); // flgs0̂Ƃ͕\ȗ.
			printf(":\n");
			cc = 0;
			for (i = 0; i < siz; i++) {
				c = fgetc(fp);
				if (c == EOF) break;
				printf("%02x ", c);
				if (((i & 15) == 15 || i + 1 == siz) && siz != 4)
					putchar('\n');
				cc = ((cc >> 8) & 0xffffff) | c << 24;
			}
			if (siz == 4)
				printf(" (=%d)\n", cc);
		}
		fclose(fp);
	}

	if (strcmp(argv[1], "string") == 0 && argc >= 4) {
		FILE *fp = openSkipHeader(argv[2], NULL, 0);
		for (;;) {
			searchId(fp, argv[3], &siz, &flg);
			if (siz == 0) break;
			fnd++;
			cc = 1;
			for (i = 0; i < siz; i++) {
				c = fgetc(fp);
				if (c == EOF) break;
				if (c == '\0') cc = 0;
				if (c < ' ') c = '.';
				if (cc != 0)
					putchar(c);
			}
			putchar('\n');
		}
		fclose(fp);
		if (fnd == 0)
			puts("not found");
	}

	if (strcmp(argv[1], "getint") == 0 && argc >= 4) {
		FILE *fp = openSkipHeader(argv[2], NULL, 0);
		for (;;) {
			searchId(fp, argv[3], &siz, &flg);
			if (siz == 0) break;
			fnd++;
			if (siz != 4) {
				puts("invalid size");
				for (i = 0; i < siz; i++)
					c = fgetc(fp);
			} else {
				fread(&i, 1, 4, fp);
				printf("%d\n", i);
			}
		}
		fclose(fp);
		if (fnd == 0)
			puts("not found");
	}

	if (strcmp(argv[1], "delete") == 0 && argc >= 4) {
		FILE *fpw = fopen("pdbtol.tmp", "wb");
		if (deleteId(argv[2], fpw, argv[3]) == 0) {
			puts("not found");
			fclose(fpw);
			exit(0);
		}
		renameTmp(fpw, "pdbtol.tmp", argv[2]);
	}

	if (strcmp(argv[1], "rename") == 0 && argc >= 5) {
		FILE *fpw = fopen("pdbtol.tmp", "wb");
		if (deleteId(argv[2], fpw, argv[3]) == 0) {
			puts("not found");
			fclose(fpw);
			exit(0);
		}
		FILE *fpr = openSkipHeader(argv[2], NULL, 0);
		searchId(fpr, argv[3], &siz, &flg);
		putSizFlgId(fpw, siz, flg, argv[4]);
		copyBytes(fpr, fpw, siz);
		renameTmp(fpw, "pdbtol.tmp", argv[2]);
	}

	if (strcmp(argv[1], "setint") == 0 && argc >= 5) {
		FILE *fpr = openSkipHeader(argv[2], NULL, 0);
		searchId(fpr, argv[3], &siz, &flg);
		fclose(fpr);
		if (siz == 0) {
			puts("not found");
			exit(0);
		}
		if (siz != 4) {
			puts("not integer");
			exit(0);
		}
		FILE *fpw = fopen("pdbtol.tmp", "wb");
		deleteId(argv[2], fpw, argv[3]);
		putSizFlgId(fpw, 4, flg, argv[3]);
		i = 0;
		sscanf(argv[4], "%d", &i);
		fwrite(&i, 1, 4, fpw);
		renameTmp(fpw, "pdbtol.tmp", argv[2]);
	}

	if (strcmp(argv[1], "new") == 0 && argc >= 5) {
		FILE *fpr = openSkipHeader(argv[2], NULL, 0);
		searchId(fpr, argv[3], &siz, &flg);
		fclose(fpr);
		if (siz != 0) {
			puts("already exist");
			exit(0);
		}
		siz = 0;
		sscanf(argv[4], "%d", &siz);
		if (siz > 0) {
			FILE *fpw = fopen("pdbtol.tmp", "wb");
			deleteId(argv[2], fpw, argv[3]);
			putSizFlgId(fpw, siz, 0, argv[3]);
			for (i = 0; i < siz; i++)
				fputc(0, fpw);
			renameTmp(fpw, "pdbtol.tmp", argv[2]);
		}
	}

	if (strcmp(argv[1], "resize") == 0 && argc >= 5) {
		FILE *fpr = openSkipHeader(argv[2], NULL, 0);
		searchId(fpr, argv[3], &siz, &flg);
		fclose(fpr);
		if (siz == 0) {
			puts("not found");
			exit(0);
		}
		FILE *fpw = fopen("pdbtol.tmp", "wb");
		deleteId(argv[2], fpw, argv[3]);
		i = 0;
		sscanf(argv[4], "%d", &i);
		if (i > 0) {
			fpr = openSkipHeader(argv[2], NULL, 0);
			searchId(fpr, argv[3], &siz, &flg);
			putSizFlgId(fpw, i, flg, argv[3]);
			if (i <= siz)
				copyBytes(fpr, fpw, i);
			else {
				copyBytes(fpr, fpw, siz);
				siz = i - siz;
				for (i = 0; i < siz; i++)
					fputc(0, fpw);
			}
			fclose(fpr);
			renameTmp(fpw, "pdbtol.tmp", argv[2]);
		}
	}

	return 0;
}

